@@ -1,21 +1,22 @@
+from __future__ import annotations
+
 import distutils.util
 import importlib
 import inspect
 import shlex
 from typing import Iterable
 
+from jb_declarative_formatters.parsers.cpp_parser import CppParser
 from jb_declarative_formatters.parsers.type_name_parser import parse_type_name_template
-from jb_declarative_formatters.type_viz_synthetic_method import SyntheticMethodDefinition
-from jb_lldb_polyfills import \
-    LLDBRemoveAllTopLevelLazyDeclarations, \
-    LLDBAddTopLevelLazyDeclarationByRegex, \
-    LLDBAddTopLevelLazyDeclaration
+from jb_declarative_formatters.type_viz_generated_method import GeneratedMethodDefinition
 from renderers.jb_lldb_builtin_formatters import *
 from renderers.jb_lldb_declarative_formatters_loaders import *
 from renderers.jb_lldb_declarative_formatters_manager import *
 from renderers.jb_lldb_format import update_value_dynamic_state
 from renderers.jb_lldb_logging import get_suppress_errors
-from renderers.jb_lldb_natvis_formatters import NatVisDescriptor
+from renderers.jb_lldb_natvis_formatters import NatVisDescriptor, NatVisSyntheticItemDescriptor
+from renderers.jb_lldb_natvis_synthetic_item_type_viz_cache import NatvisSyntheticItemTypeVizCache
+from renderers.jb_lldb_top_level_lazy_declarations import LLDBTopLevelLazyDeclarations
 
 lldb_formatters_manager: FormattersManager
 
@@ -53,8 +54,6 @@ def __lldb_init_module(debugger: lldb.SBDebugger, internal_dict):
     viz_provider = VizDescriptorProvider()
     set_viz_descriptor_provider(viz_provider)
 
-    start_listener(debugger)
-
 
 def _cmd_loaders_add(debugger, command, exe_ctx, result, internal_dict):
     # raise NotImplementedError("jb_renderers_loaders_add is not implemented yet")
@@ -228,44 +227,27 @@ def get_all_registered_files():
 def remove_file_list(debugger, files):
     for filepath in files:
         lldb_formatters_manager.unregister(filepath)
-    LLDBRemoveAllTopLevelLazyDeclarations(debugger)
+    NatvisSyntheticItemTypeVizCache.clear_cache()
+    LLDBTopLevelLazyDeclarations.remove_all_top_level_lazy_declarations(debugger)
     add_all_top_level_declarations(debugger, lldb_formatters_manager.formatter_entries.values())
 
 
 def reload_file_list(debugger, files):
     for filepath in files:
         lldb_formatters_manager.reload(filepath)
-    LLDBRemoveAllTopLevelLazyDeclarations(debugger)
+    NatvisSyntheticItemTypeVizCache.clear_cache()
+    LLDBTopLevelLazyDeclarations.remove_all_top_level_lazy_declarations(debugger)
     add_all_top_level_declarations(debugger, lldb_formatters_manager.formatter_entries.values())
 
 
 def add_all_top_level_declarations(debugger: lldb.SBDebugger, entries: Iterable[FormattersManager.FormatterEntry]):
     for entry in entries:
-        top_level_methods: List[SyntheticMethodDefinition] = entry.storage.get_top_level_methods()
+        top_level_methods: list[GeneratedMethodDefinition] = entry.storage.get_top_level_methods()
         for top_level_method_definition in top_level_methods:
-            if top_level_method_definition.name_uses_regex:
-                error: lldb.SBError = LLDBAddTopLevelLazyDeclarationByRegex(
-                    debugger,
-                    top_level_method_definition.full_name,
-                    top_level_method_definition.body_substitution,
-                    eLanguageTypeC_plus_plus_14
-                )
-                if not error.Success():
-                    log(f"Can't add lazy declarations by regex for '{top_level_method_definition.full_name}': "
-                        f"{error.description}")
-            else:
-                error: lldb.SBError = LLDBAddTopLevelLazyDeclaration(
-                    debugger,
-                    top_level_method_definition.full_name,
-                    top_level_method_definition.body_substitution,
-                    eLanguageTypeC_plus_plus_14
-                )
-                if not error.Success():
-                    log(f"Can't add lazy declarations for '{top_level_method_definition.full_name}': "
-                        f"{error.description}")
+            LLDBTopLevelLazyDeclarations.declare_lazy_declaration(debugger, top_level_method_definition)
 
 
-def declarative_summary(val: lldb.SBValue, internal_dict):
+def declarative_summary(val: lldb.SBValue, _):
     try:
         update_value_dynamic_state(val)
         val_non_synth = val.GetNonSyntheticValue()
@@ -286,27 +268,100 @@ def declarative_summary(val: lldb.SBValue, internal_dict):
 
 
 class DeclarativeSynthProvider(object):
-    def __init__(self, val, internal_dict):
+    """
+    Implementation of SyntheticChildrenProvider from LLDB: https://lldb.llvm.org/use/variable.html#synthetic-children
+    """
+
+    def __init__(self, val: lldb.SBValue, _):
+        """
+        This call should initialize the Python object using val as the variable to provide synthetic children for.
+        """
         update_value_dynamic_state(val)
         self.val_non_synth: lldb.SBValue = val.GetNonSyntheticValue()
         self.children_provider: Optional[AbstractChildrenProvider] = None
 
-    def update(self):
+    def num_children(self, max_children: int) -> int:
+        """
+        This call should return the number of children that you want your object to have.
+
+        :param max_children: The max_children argument indicates the maximum number of children that lldb is interested in (at this moment).
+        If the computation of the number of children is expensive (for example, requires traversing a linked list to determine its size)
+        your implementation may return `max_children` rather than the actual number. If the computation is cheap (e.g., the number is stored
+        as a field of the object), then you can always return the true number of children (that is, ignore the `max_children` argument).
+        """
+        if not self.children_provider:
+            self._create_children_provider()
+        else:
+            self.children_provider.try_update_size(self.val_non_synth)
+        return self.children_provider.num_children()
+
+    def get_child_index(self, name: str) -> int:
+        """
+        This call should return the index of the synthetic child whose name is given as argument.
+        """
+        # Ensure that `children_provider` is created because there is API which can call this method without prior call to `num_children`:
+        # GetChildMemberWithName, GetIndexOfChildWithName
+        if not self.children_provider:
+            self._create_children_provider()
+        return self.children_provider.get_child_index(name)
+
+    def get_child_at_index(self, index: int) -> lldb.SBValue:
+        """
+        This call should return a new LLDB SBValue object representing the child at the index given as argument.
+        """
+        # Ensure that `children_provider` is created because there is API which can call this method without prior call to `num_children`:
+        # GetChildMemberWithName, GetChildAtIndex
+        if not self.children_provider:
+            self._create_children_provider()
+        return self.children_provider.get_child_at_index(index)
+
+    # noinspection PyMethodMayBeStatic
+    def update(self) -> bool:
+        """
+        This call should be used to update the internal state of this Python object whenever the state of the variables in LLDB changes.
+        Also, this method is invoked before any other method in the interface.
+
+        :return: If `False` is returned, then whenever the process reaches a new stop, this method will be invoked again to generate
+        an updated list of the children for a given variable. Otherwise, if `True` is returned, then the value is cached and this method
+        won’t be called again, effectively freezing the state of the value in subsequent stops.
+        Beware that returning `True` incorrectly could show misleading information to the user.
+
+        P.S.
+        In the LLDB code there is following comment which actually contradicts to the official documentation:
+            // this function is assumed to always succeed and it if fails, the front-end
+            // should know to deal with it in the correct way (most probably, by refusing
+            // to return any children) the return value of Update() should actually be
+            // interpreted as "ValueObjectSyntheticFilter cache is good/bad" if =true,
+            // ValueObjectSyntheticFilter is allowed to use the children it fetched
+            // previously and cached if =false, ValueObjectSyntheticFilter must throw
+            // away its cache, and query again for children
+        While tests will continue work correctly if returning `True` here, it doesn't improve their execution time.
+        """
+        # We do not create `children_provider` here because it might be not needed later so why waste CPU time.
+        # Also, we do not call `children_provider.try_update_size` because this `update` is called once for each child!
+        # We rely on the fact that before enumerating children LLDB will call `num_children` and we call `try_update_size` from there.
         return False
 
-    def has_children(self):
+    def has_children(self) -> bool:
+        """
+        This call should return `True` if this object might have children, and `False` if this object can be guaranteed to have no children.
+        """
         return self.val_non_synth.MightHaveChildren()
 
-    def ensure_initialized(self):
-        if self.children_provider:
-            return
+    # def get_value(self) -> lldb.SBValue:
+    #     """
+    #     This call can return an `SBValue` to be presented as the value of the synthetic value under consideration.
+    #     The `SBValue` you return here will most likely be a numeric type (int, float, …) as its value bytes
+    #     will be used as-if they were the value of the root SBValue proper.
+    #     """
+    #     pass
+
+    def _create_children_provider(self) -> None:
         try:
             log("Retrieving children of value named '{}'...", self.val_non_synth.GetName())
 
-            format_spec = self.val_non_synth.GetFormat()
-            use_raw_viz = format_spec & eFormatRawView
             provider = get_viz_descriptor_provider()
-            vis_descriptor = provider.get_matched_visualizers(self.val_non_synth.GetType(), use_raw_viz)
+            vis_descriptor = provider.get_matched_visualizers(self.val_non_synth, False)
             if vis_descriptor:
                 self.children_provider = vis_descriptor.prepare_children(self.val_non_synth)
 
@@ -320,40 +375,25 @@ class DeclarativeSynthProvider(object):
         if not self.children_provider:
             self.children_provider = StructChildrenProvider(self.val_non_synth)
 
-    def num_children(self):
-        self.ensure_initialized()
-        return self.children_provider.num_children()
-
-    def get_child_index(self, name):
-        self.ensure_initialized()
-        return self.children_provider.get_child_index(name)
-
-    def get_child_at_index(self, index):
-        self.ensure_initialized()
-        return self.children_provider.get_child_at_index(index)
-
 
 class VizDescriptorProvider(AbstractVizDescriptorProvider):
     def __init__(self):
         self.type_to_visualizer_cache = {}
-        self.type_to_raw_view_visualizer_cache = {}
 
-    def get_matched_visualizers(self, value_type: lldb.SBType, raw_visualizer: bool) -> AbstractVisDescriptor:
-        if raw_visualizer:
-            cache = self.type_to_raw_view_visualizer_cache
-            use_natvis = False
-        else:
-            cache = self.type_to_visualizer_cache
-            use_natvis = True
-
-        type_name = value_type.GetName()
-
-        descriptor = cache.get(type_name, None)
+    def get_matched_visualizers(self, val_non_synth: lldb.SBValue, force_raw_format: bool) -> AbstractVisDescriptor:
+        descriptor = _try_get_natvis_synthetic_item_visualizers(val_non_synth)
         if descriptor is not None:
             return descriptor
 
-        descriptor = _try_get_matched_visualizers(value_type, use_natvis)
-        cache[type_name] = descriptor
+        format_spec = eFormatRawView if force_raw_format else val_non_synth.GetFormat()
+        value_type = val_non_synth.GetType()
+        cache_key = (value_type.GetName(), format_spec)
+        descriptor = self.type_to_visualizer_cache.get(cache_key, None)
+        if descriptor is not None:
+            return descriptor
+
+        descriptor = _try_get_matched_visualizers(value_type, format_spec)
+        self.type_to_visualizer_cache[cache_key] = descriptor
 
         return descriptor
 
@@ -393,11 +433,11 @@ def _try_find_matched_natvis_visualizer_for_base(value_type: lldb.SBType) -> Opt
     return None
 
 
-def _try_get_matched_visualizers(value_type: lldb.SBType, natvis_enabled) -> Optional[AbstractVisDescriptor]:
+def _try_get_matched_visualizers(value_type: lldb.SBType, format_spec: int) -> Optional[AbstractVisDescriptor]:
     value_type: lldb.SBType = value_type.GetUnqualifiedType()
-    value_type_name = value_type.GetName()
 
-    if natvis_enabled:
+    if not (format_spec & eFormatRawView):
+        value_type_name = CppParser.remove_type_class_specifier(value_type.GetName())
         log("Trying to find natvis visualizer for type: '{}'...", value_type_name)
         try:
             type_name_template = parse_type_name_template(value_type_name)
@@ -409,10 +449,10 @@ def _try_get_matched_visualizers(value_type: lldb.SBType, natvis_enabled) -> Opt
             log("Found natvis visualizer for type: '{}'", value_type_name)
             return NatVisDescriptor(viz_candidates, type_name_template)
 
-    return _try_get_matched_builtin_visualizer(value_type, natvis_enabled)
+    return _try_get_matched_builtin_visualizer(value_type, format_spec)
 
 
-def _try_get_matched_builtin_visualizer(value_type, natvis_enabled):
+def _try_get_matched_builtin_visualizer(value_type: lldb.SBType, format_spec: int):
     value_type_name = value_type.GetName()
     log("Trying to find builtin visualizer for type: '{}'", value_type_name)
 
@@ -422,25 +462,37 @@ def _try_get_matched_builtin_visualizer(value_type, natvis_enabled):
         value_typedef_type_name = value_typedef_type.GetName()
         log("Type '{}' is typedef to type '{}'", value_type_name, value_typedef_type_name)
         if value_typedef_type_name != value_type_name:
-            return _try_get_matched_visualizers(value_typedef_type, natvis_enabled)
+            return _try_get_matched_visualizers(value_typedef_type, format_spec)
 
+    numeric_value_only = format_spec & eFormatNumericValueOnly
     if type_class == lldb.eTypeClassBuiltin:
-        char_presentation_info = CharVisDescriptor.char_types.get(value_type_name)
-        if char_presentation_info is not None:
-            return CharVisDescriptor(char_presentation_info)
-        if value_type_name in NumberVisDescriptor.numeric_types:
+        basic_fmt_spec = format_spec & eFormatBasicSpecsMask
+        str_presentation_info = FMT_STRING_SET_ALL.get(basic_fmt_spec)
+        # When a format specifier is used, VS implicitly converts some integer types to a string pointer
+        if str_presentation_info is not None and CharArrayOrPointerVisDescriptor.can_type_be_used_as_char_pointer(value_type):
+            return CharArrayOrPointerVisDescriptor(str_presentation_info, False, None)
+
+        if not numeric_value_only:
+            char_presentation_info = CharVisDescriptor.char_types.get(value_type_name)
+            if char_presentation_info is not None:
+                return CharVisDescriptor(char_presentation_info)
+        if NumberVisDescriptor.is_number_type(value_type_name):
             return NumberVisDescriptor(value_type_name)
 
     if type_class == lldb.eTypeClassArray:
-        array_element_type: SBType = value_type.GetArrayElementType()
-        array_element_type_name = array_element_type.GetName()
-        str_presentation_info = CharVisDescriptor.char_types.get(array_element_type_name)
-        if str_presentation_info is not None:
-            array_size = value_type.size // array_element_type.GetByteSize()
-            return CharArrayOrPointerVisDescriptor(str_presentation_info, True, array_size)
+        if not numeric_value_only:
+            array_element_type: SBType = value_type.GetArrayElementType()
+            array_element_type_name = array_element_type.GetName()
+            str_presentation_info = CharVisDescriptor.char_types.get(array_element_type_name)
+            if str_presentation_info is not None:
+                array_size = value_type.size // array_element_type.GetByteSize()
+                return CharArrayOrPointerVisDescriptor(str_presentation_info, True, array_size)
         return GenericArrayVisDescriptor()
 
     if type_class == lldb.eTypeClassPointer:
+        if numeric_value_only:
+            return PointerAsIntegerVisDescriptor(value_type.GetByteSize())
+
         pointee_type: SBType = value_type.GetPointeeType()
         pointee_type_name = pointee_type.GetName()
         str_presentation_info = CharVisDescriptor.char_types.get(pointee_type_name)
@@ -460,11 +512,11 @@ def _try_get_matched_builtin_visualizer(value_type, natvis_enabled):
         return GenericReferenceVisDescriptor()
 
     if type_class == lldb.eTypeClassStruct or type_class == lldb.eTypeClassClass or type_class == lldb.eTypeClassUnion:
-        if natvis_enabled:
+        if not (format_spec & eFormatRawView):
             natvis = _try_find_matched_natvis_visualizer_for_base(value_type)
             if natvis is not None:
                 return natvis
-        lambda_name = _try_extract_lambda_type_name(value_type_name)
+        lambda_name = CppParser.try_extract_lambda_type_name(value_type_name)
         if lambda_name is not None:
             return LambdaVisDescriptor(value_type, lambda_name)
         return StructVisDescriptor(value_type)
@@ -473,15 +525,9 @@ def _try_get_matched_builtin_visualizer(value_type, natvis_enabled):
     return None
 
 
-def _try_extract_lambda_type_name(type_name: str) -> Optional[str]:
-    idx = type_name.rfind('<lambda_')
-    if idx == -1:
+def _try_get_natvis_synthetic_item_visualizers(val_non_synth: lldb.SBValue) -> NatVisSyntheticItemDescriptor | None:
+    type_viz_synthetic_item, wildcards = NatvisSyntheticItemTypeVizCache.retrieve_type_viz_and_wildcards_from(val_non_synth)
+    if not type_viz_synthetic_item:
         return None
-    if type_name[-1] != '>':
-        return None
-    if idx == 0:
-        return type_name
-    extracted_name = type_name[idx + len("<lambda_"):-1]
-    if not extracted_name.isalnum():
-        return None
-    return type_name[idx:]
+
+    return NatVisSyntheticItemDescriptor(type_viz_synthetic_item, wildcards)
